"Python Databse Wrapper class and DB Factory"
import logging
import sys
from typing import Dict, List, NewType, Tuple, Union

# import mariadb
# import cx_Oracle
import mysql.connector

# import snowflake.connector

DB_ROWS = NewType("DB_RET", List[Tuple])
DB_INFO = Union[Dict[str, int], Dict[str, str]]

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
fmt = "[ %(asctime)s ] %(levelname)s|%(name)s|%(message)s"
handler.setFormatter(fmt)
logger.addHandler(handler)


def convert_query_result(query_res):
    """Return List(elements) from query result List(Tuples(element))."""
    if len(query_res) == 0:
        logger.warning("Received empty result set from DB")
        return []
    return [list(i) for i in query_res] if len(query_res) > 1 else query_res[0][0]


class DbWrapper:
    """Base Class for DB Connection Wrapper"""

    def __init__(self, info: DB_INFO, connector, exception=None, dictionary=False):
        """DB Connection Wrapper Base Class Constructor

        Args:
            info (DB_INFO [dict]): DB Connection Info
            connector (connection function pointer): Function pointer to the connect function
            exception (Exception, optional): Exception for the DB error handling. Defaults to base Exception.
        """
        try:
            self._conn = connector(**info)
            if dictionary:
                self._dictionary = True
                self._cur = self._conn.cursor(dictionary=True)
            else:
                self._dictionary = False
                self._cur = self._conn.cursor()
            self._exception = exception if exception is not None else Exception
        except self._exception as e:
            raise e

    def get_connection(self):
        """Returns the connection object for the DB"""
        return self._conn

    def get_cursor(self):
        """Returns a reference to the cursor object

        Returns:
            cursor: the cursor object for the db
        """
        return self._cur

    def get_exception(self):
        """Returns the exception handler for the class

        Returns:
            [type]: [description]
        """
        return self._exception

    def query(self, stmt):
        """Queries the db with <stmt>.  Returns list of tuples if there are results."""
        try:
            self._cur.execute(stmt)
            res = self._cur.fetchall()
            return convert_query_result(res) if not self._dictionary else res
        except self._exception as exception:
            raise exception

    def query_with_commit(self, stmt):
        """Queries the db with <stmt>.  Returns list of tuples if there are results."""
        try:
            self._cur.execute(stmt)
            self._cur.commit()
        except self._exception as exception:
            raise exception

    def execute(self, stmt):
        """Exectes a query, commits, and resturns result."""
        try:
            res = self._cur.execute(stmt)
            self._conn.commit()
            return (
                convert_query_result(res.fetchall())
                if not self._dictionary
                else self._cur.fetchall()
            )
        except self._exception as e:
            raise e

    def close(self):
        """Close the db connection and cursor."""
        self._cur.close()
        self._conn.close()


class MysqlDB(DbWrapper):
    """Mysql Specific Functions"""

    def __init__(self, info, dictionary=False):
        """MySQL Connection Wrapper"""
        DbWrapper.__init__(
            self,
            info,
            mysql.connector.connect,
            mysql.connector.Error,
            dictionary=dictionary,
        )
        self._dictionary = dictionary
        self._conn = self.get_connection()
        self._cur = self.get_cursor()

    def query_with_params(self, stmt, params):
        """Queries db with with <stmt> and <params>."""
        try:
            self._cur.execute(stmt, params)
            return (
                convert_query_result(self._cur.fetchall())
                if not self._dictionary
                else self._cur.fetchall()
            )
        except self.get_exception() as exception:
            raise exception

    def query_with_params_and_commit(self, stmt, params):
        """Queries the db with <stmt> and <params> and commits."""
        try:
            self._cur.execute(stmt, params)
            self._conn.commit()
        except self.get_exception() as exception:
            raise exception

    def get_curdate(self):
        """Returns CURDATE() from MySQL."""
        return self.query("SELECT CURDATE()")[0]

    def get_timestamp(self):
        """Returns CURRENT_TIMESTAMP from MySQL."""
        return self.query("SELECT CURRENT_TIMESTAMP()")[0]

    def table_exists(self, schema: str, table: str):
        stmt = f"""
        SELECT COUNT(*) from information_schema.TABLES
        WHERE UPPER(TABLE_SCHEMA) = UPPER('{schema}')
        AND UPPER(TABLE_NAME) = UPPER('{table}')
        """
        return self.query(stmt)[0][0] != 0


# class SnowflakeWrapper(DB_WRAPPER):
#     """Snowflake Specific Functions"""

#     def __init__(self, info: DB_INFO, connector, exception):
#         DB_WRAPPER.__init__(self, info, connector, exception)


# class OracleWrapper(DB_WRAPPER):
#     """Oracle specific functions."""

#     def __init__(self, info: DB_INFO, connector, exception):
#         DB_WRAPPER.__init__(self, info, connector, exception)

#     def get_incoming_me(self):
#         """Returns the ME_INCOMING table from Oracle."""
#         return self.query(
#             "SELECT post_day_end.pde_commons_pkg.get_incoming_daybreak_me_name FROM dual"
#         )[0]


class DatabaseManager:
    """Context Manager for DB Connection"""

    def __init__(self, db_info: DB_INFO, db_type: str, dictionary=False):
        self._db_type = db_type
        self._db_info = db_info
        self._dictionary = dictionary
        self._db = db_factory(self._db_info, self._db_type, dictionary=self._dictionary)

    def __enter__(self):
        return self._db

    def __exit__(self, exc_type, exc_value, traceback):
        self._db.close()


def db_factory(
    db_info: Union[Dict[str, int], Dict[str, str]], db_type: str, dictionary=False
):
    """Build the correct db wrapper for the db_type."""
    db_type = db_type.strip().lower()
    if db_type == "mysql":
        dbo = MysqlDB(db_info, dictionary=dictionary)
    else:
        logger.error("ERROR %s not valid", db_type)
        logger.error("Valid types: [ mysql ]")
        sys.exit(1)
    return dbo
    # elif db_type == "snowflake":
    #     return SnowflakeWrapper(
    #         db_info, snowflake.connector.connect, snowflake.connector.Error
    #     )
    # elif db_type in ("oracle", "prepdb", "bengal", "livdb", "slivdb"):
    #     return OracleWrapper(
    #         db_info,
    #         cx_Oracle.connect,
    #         cx_Oracle.DatabaseError,
    #     )
    # elif db_type == 'mariadb':
    #     return _MariaDB(db_info)
